/*
 * @(#)ExplicitHConstraints.java
 * 
 * Christopher Deckers (chrriis@brainlex.com)
 * http://chrriis.brainlex.com
 * 
 * See the file "LICENSE.txt" for information on usage and redistribution of
 * this file, and for a DISCLAIMER OF ALL WARRANTIES.
 */
package chrriis.uihierarchy.constraints;

import java.awt.Component;
import java.awt.Container;

import com.zookitec.layout.ComponentEF;
import com.zookitec.layout.ContainerEF;
import com.zookitec.layout.ExplicitConstraints;
import com.zookitec.layout.ExplicitLayout;
import com.zookitec.layout.Expression;
import com.zookitec.layout.GroupEF;
import com.zookitec.layout.MathEF;

import chrriis.uihierarchy.Accessor;
import chrriis.uihierarchy.SingleLevelAccessor;
import chrriis.uihierarchy.util.HStringTokenizer;
import chrriis.uihierarchy.util.UnitParser;

/**
 * The constraints for an Explicit layout (www.zookitec.com). Constraints are
 * manipulated by the <code>createConstraints</code> method, if they are a
 * String.
 * <br><br>
 * The format is of the form "inherit, key1=expression1, key2=expression2,
 * ...". Inherit is optional and can have two forms: "inherit" or
 * "inherit=mappedName". The inherit keyword is used to first apply some
 * earlier constraints before applying a new set of constraints. This allows
 * chaining of constraints. If used with a mapped name then it will use the
 * constraints applied to the level that was mapped, else it will use the
 * previous sibbling's constraints.<br>
 * <br>
 * Possible keys are (case is insensitive):<br>
 * x, y, originX, originY, bottom, right, width and height.<br>
 * <br>
 * The expressions are combinations of constants, component-based constraints,
 * and functions. An example can be: "5+comp.top*max(3, 4)", where "comp" is
 * a component that was mapped to this name.<br>
 * The word "container", if not mapped, can be used to refer to the parent
 * container.<br>
 * <br>
 * Constants are double values. Possible forms are integer form (e.g. "5") or
 * double form (e.g. "5.0").<br>
 * Predefined supported constants are: bottom, center, left, right and top.
 * <br><br>
 * Component-based constraints are:<br>
 * For component, container and group: bottom, right, width, height, left,
 * top.<br>
 * For component only: minHeight, minWidth, maxHeight, maxWidth, prefHeight,
 * prefWidth.<br>
 * For group only: heightMax, heightMin, heightSum, widthMax, widthMin,
 * widthSum, prefHeightMax, prefHeightMin, prefHeightSum, prefWidthMax,
 * prefWidthMin and prefWidthSum.<br>
 * There are two possibilities to use a group: the mapped name matches several
 * components, or use a form like "[mappedName1, mappedName2, ...].top".<br>
 * Additionaly, there is support for the <code>xFraction()</code> and
 * <code>yFraction()</code> methods, by using "mappedName.x" or "mappedName.y"
 * followed by a division or a multiplication. An example could be
 * "x=container.x/2".<br>
 * <br>
 * Supported functions are:<br>
 * min(exp1, exp2, ...), max(exp1, exp2, ...), bound(exp1, exp2, exp3),
 * gridX(int, exp) and gridY(int, exp).<br>
 * <br>
 * The properties <code>heightZeroIfInvisible</code> and
 * <code>widthZeroIfInvisible</code> can be set by using "= true" or
 * "= false".
 * @version 1.1 2003.11.03
 * @author Christopher Deckers (chrriis@brainlex.com)
 */
public class ExplicitHConstraints implements LayoutHConstraints {

  /**
   * Create the constraints for an Explicit layout, if the parameter is a
   * String.
   * @param accessor The accessor of the level for which to create the
   * constraints.
   * @param parentContainer The parent container.
   * @param constraints The constraints of the component.
   * @return The ExplicitConstraints constraints.
   * @see chrriis.uihierarchy.HLeaf#map(java.lang.String)
   * @see chrriis.uihierarchy.HNode#map(java.lang.String)
   * @see chrriis.uihierarchy.HRootNode#map(java.lang.String)
   */
  public Object createConstraints(SingleLevelAccessor accessor, Container parentContainer, Object constraints) {
    if(!(constraints instanceof String)) {
      return constraints;
    }
    HStringTokenizer[] tokenizers = new HStringTokenizer((String)constraints).getSubTokenizers();
    ExplicitConstraints explicitConstraints = new ExplicitConstraints(accessor.getComponent());
    ExplicitLayout layout = (ExplicitLayout)parentContainer.getLayout();
    for(int i=0; i<tokenizers.length; i++) {
      HStringTokenizer st = tokenizers[i];
      char operator = st.getOperator();
      HStringTokenizer[] subTokenizers = st.getSubTokenizers();
      if(operator != '=') {
        String function = st.getToken().toLowerCase();
        if(function.equals("inherit")) {
          if(i==0) {
            try {
              SingleLevelAccessor valueAccessor = accessor.previous();
              String sConstraints = (String)constraints;
              int index = sConstraints.indexOf(",");
              if(index >= 0) {
                constraints = valueAccessor.getConstraintsHandler().getRealizedConstraints() + sConstraints.substring(index);
              } else {
                constraints = valueAccessor.getConstraintsHandler().getRealizedConstraints();
              }
              explicitConstraints.copy((ExplicitConstraints)createConstraints(accessor, parentContainer, constraints));
              break;
            } catch(Exception e) {
              String message = e.getMessage();
              message = message != null? " " + message: "";
              throw new IllegalArgumentException("Failed to \"inherit\"." + message);
            }
          }
          throw new IllegalArgumentException("The \"inherit\" constraint can only be used as a first constraint!");
        }
        throw new IllegalStateException("Constraints must be key/expression pairs!");
      }
      String left = subTokenizers[0].getToken().toLowerCase();
      if(left.equals("inherit")) {
        if(i != 0) {
          throw new IllegalArgumentException("The \"inherit\" constraint can only be used as a first constraint!");
        }
        try {
          String right = subTokenizers[1].getToken();
          if(right == null) {
            right = "";
            HStringTokenizer[] sts = subTokenizers[1].getSubTokenizers();
            for(int j=0; j<sts.length; j++) {
              right+="," + sts[j].getToken();
            }
          }
          SingleLevelAccessor valueAccessor = (SingleLevelAccessor)accessor.getConfig().createAccessor(right);
          String sConstraints = (String)constraints;
          int index = sConstraints.indexOf(",");
          if(index >= 0) {
            constraints = valueAccessor.getConstraintsHandler().getRealizedConstraints() + sConstraints.substring(index);
          } else {
            constraints = valueAccessor.getConstraintsHandler().getRealizedConstraints();
          }
          explicitConstraints.copy((ExplicitConstraints)createConstraints(accessor, parentContainer, constraints));
        } catch(Exception e) {
          String message = e.getMessage();
          message = message != null? " " + message: "";
          throw new IllegalArgumentException("Failed to \"inherit\"." + message);
        }
      } else if(left.equals("heightzeroifinvisible") || left.equals("widthzeroifinvisible")) {
        String right = subTokenizers[1].getToken().toLowerCase();
        if(right.equals("true") || right.equals("false")) {
          if(left.equals("heightzeroifinvisible")) {
            explicitConstraints.setHeightZeroIfInvisible(Boolean.valueOf(right).booleanValue());
          } else {
            explicitConstraints.setWidthZeroIfInvisible(Boolean.valueOf(right).booleanValue());
          }
        } else {
          throw new IllegalStateException("\"" + left + "\" must have a parameter equal to \"true\" or \"false\"!");
        }
      } else if(left.equals("x")) {
        explicitConstraints.setX(createExpression(accessor, parentContainer, subTokenizers[1]));
      } else if(left.equals("y")) {
        explicitConstraints.setY(createExpression(accessor, parentContainer, subTokenizers[1]));
      } else if(left.equals("originx")) {
        explicitConstraints.setOriginX(createExpression(accessor, parentContainer, subTokenizers[1]).getValue(layout));
      } else if(left.equals("originy")) {
        explicitConstraints.setOriginY(createExpression(accessor, parentContainer, subTokenizers[1]).getValue(layout));
      } else if(left.equals("bottom")) {
        explicitConstraints.setBottom(createExpression(accessor, parentContainer, subTokenizers[1]));
      } else if(left.equals("right")) {
        explicitConstraints.setRight(createExpression(accessor, parentContainer, subTokenizers[1]));
      } else if(left.equals("width")) {
        explicitConstraints.setWidth(createExpression(accessor, parentContainer, subTokenizers[1]));
      } else if(left.equals("height")) {
        explicitConstraints.setHeight(createExpression(accessor, parentContainer, subTokenizers[1]));
      } else {
        throw new IllegalStateException("\"" + subTokenizers[0].getToken() + "\" is not a valid key!");
      }
    }
    accessor.getConstraintsHandler().setRealizedConstraints(constraints);
//    accessor.getConstraintsHandler().setRealizedConstraints(explicitConstraints.clone());
    return explicitConstraints;
  }

  /**
   * Create an expression from a hierarchy tokenizer representation.
   * @param accessor The accessor of the level for which to create the
   * constraints.
   * @param tokenizer The hierarchy tokenizer representing the constraints.
   * @return The expression created from the tokenizer.
   */
  protected Expression createExpression(SingleLevelAccessor accessor, Container parentContainer, HStringTokenizer tokenizer) {
    char operator = tokenizer.getOperator();
    String token = tokenizer.getToken();
    HStringTokenizer[] sts = tokenizer.getSubTokenizers();
    switch(operator) {
      case '-':
        if(sts[0] == null) {
          return MathEF.subtract(MathEF.constant(0), createExpression(accessor, parentContainer, sts[1]));
        }
        return MathEF.subtract(createExpression(accessor, parentContainer, sts[0]), createExpression(accessor, parentContainer, sts[1]));
      case '+':
        return MathEF.add(createExpression(accessor, parentContainer, sts[0]), createExpression(accessor, parentContainer, sts[1]));
      case '*':
        String leftToken = sts[0].getToken();
        Expression multiplicationExpression = MathEF.multiply(createExpression(accessor, parentContainer, sts[0]), createExpression(accessor, parentContainer, sts[1]));
        try {
          if(leftToken != null) {
            if(leftToken.endsWith(".x")) {
              return MathEF.add(getStringExpression(accessor, parentContainer, leftToken.substring(0, leftToken.length() - 2) + ".left"), multiplicationExpression);
            } else if(leftToken.endsWith(".y")) {
              return MathEF.add(getStringExpression(accessor, parentContainer, leftToken.substring(0, leftToken.length() - 2) + ".top"), multiplicationExpression);
            }
          }
        } catch(Exception e){}
        return multiplicationExpression;
      case '/':
        leftToken = sts[0].getToken();
        Expression divisionExpression = MathEF.divide(createExpression(accessor, parentContainer, sts[0]), createExpression(accessor, parentContainer, sts[1]));
        try {
          if(leftToken != null) {
            if(leftToken.endsWith(".x")) {
              return MathEF.add(getStringExpression(accessor, parentContainer, leftToken.substring(0, leftToken.length() - 2) + ".left"), divisionExpression);
            } else if(leftToken.endsWith(".y")) {
              return MathEF.add(getStringExpression(accessor, parentContainer, leftToken.substring(0, leftToken.length() - 2) + ".top"), divisionExpression);
            }
          }
        } catch(Exception e){}
        return divisionExpression;
      case '(':
        if(token != null) {
          String function = token.toLowerCase();
          if(function.equals("max") || function.equals("min")) {
            if(sts.length < 1) {
              throw new IllegalStateException("\"" + token + "\" function must have one or more parameters!");
            }
            Expression[] expression = new Expression[sts.length];
            for(int i=0; i<sts.length; i++) {
              expression[i] = createExpression(accessor, parentContainer, sts[i]);
            }
            if(function.equals("max")) {
              return MathEF.max(expression);
            }
            return MathEF.min(expression);
          }
          if(function.equals("bound")) {
            if(sts.length != 3) {
              throw new IllegalStateException("\"" + token + "\" function must have three parameters!");
            }
            return MathEF.bound(createExpression(accessor, parentContainer, sts[0]), createExpression(accessor, parentContainer, sts[1]), createExpression(accessor, parentContainer, sts[2]));
          }
          if(function.equals("gridx") || function.equals("gridy")) {
            if(sts.length != 2) {
              throw new IllegalStateException("\"" + token + "\" function must have two parameters!");
            }
            try {
              if(function.equals("gridx")) {
                return ContainerEF.gridX(parentContainer, Integer.parseInt(sts[0].getToken()), createExpression(accessor, parentContainer, sts[1]));
              }
              return ContainerEF.gridY(parentContainer, Integer.parseInt(sts[0].getToken()), createExpression(accessor, parentContainer, sts[1]));
            } catch(Exception e) {
              throw new IllegalStateException("Illegal format of constraints for \"" + token + "\" function!");
            }
          }
          throw new IllegalStateException("\"" + token + "\" is not a valid function!");
        }
        if(sts.length != 1) {
          throw new IllegalStateException("Braces must have one and only one expression!");
        }
        return createExpression(accessor, parentContainer, sts[0]);
      case 0:
        if(sts.length == 2 && sts[0].getOperator() == '[' && sts[1].getToken() != null && sts[1].getToken().startsWith(".")) {
          if(sts[0].getToken() != null) {
            throw new IllegalStateException("Function text \"" + sts[0].getToken() + "\" cannot be applied to a group of mapped levels!");
          }
          HStringTokenizer[] substs = sts[0].getSubTokenizers();
          String left = null;
          for(int i=0; i<substs.length; i++) {
            if(i == 0) {
              left = substs[i].getToken();
            } else {
              left += "," + substs[i].getToken();
            }
          }
          return getComponentExpression(accessor, parentContainer, token, left, sts[1].getToken().substring(1));
        }
        return getStringExpression(accessor, parentContainer, token);
      default:
        throw new IllegalStateException("Cannot handle operator \"" + operator + "\"!");
    }
  }

  /**
   * Get an expression from a String (not an operator).
   * @param accessor The accessor to the component being constrained.
   * @param parentContainer The parent container.
   * @param token The token that is the String.
   * @return The corresponding expression.
   */
  protected Expression getStringExpression(SingleLevelAccessor accessor, Container parentContainer, String token) {
    if(token == null) {
      throw new IllegalStateException("A group must be composed of one and only one expression!");
    }
    int index = token.indexOf(".");
    if(index > 0) {
      try {
        return MathEF.constant(UnitParser.doublePixelValue(accessor.getConfig(), token));
//        return MathEF.constant(Double.parseDouble(token));
      } catch(Exception e) {}
//      } catch(NumberFormatException e) {}
      return getComponentExpression(accessor, parentContainer, token, token.substring(0, index), token.substring(index+1));
    }
    String value = token.toLowerCase();
    if(value.equals("bottom")) {
      return MathEF.constant(ExplicitConstraints.BOTTOM);
    } else if(value.equals("center")) {
      return MathEF.constant(ExplicitConstraints.CENTER);
    } else if(value.equals("left")) {
      return MathEF.constant(ExplicitConstraints.LEFT);
    } else if(value.equals("right")) {
      return MathEF.constant(ExplicitConstraints.RIGHT);
    } else if(value.equals("top")) {
      return MathEF.constant(ExplicitConstraints.TOP);
    }
    try {
      return MathEF.constant(UnitParser.doublePixelValue(accessor.getConfig(), token));
//      return MathEF.constant(Double.parseDouble(token));
    } catch(Exception e) {
      throw new IllegalStateException("Unknown constraint \"" + token + "\"!");
    }
  }

  /**
   * Get the expression for some constraints related to components.
   * @param accessor The accessor to the component for which to create the
   * constraints.
   * @param parentContainer The parent of the container.
   * @param token The general token.
   * @param left The left part that designate a component, group or the
   * container.
   * @param right The constraint to extract.
   * @return The expression.
   */
  protected Expression getComponentExpression(SingleLevelAccessor accessor, Container parentContainer, String token, String left, String right) {
    Component[] components;
    try {
      Accessor leftAcc = accessor.getConfig().createAccessor(left);
      components = leftAcc.getComponents();
    } catch(RuntimeException e) {
      if(left.toLowerCase().equals("container")) {
        components = new Component[] {parentContainer};
      } else {
        throw e;
      }
    }
    right = right.toLowerCase();
    if(components.length > 1) {
      if(right.equals("bottom")) {
        return GroupEF.bottom(components);
      } else if(right.equals("right")) {
        return GroupEF.right(components);
      } else if(right.equals("width") || right.equals("x")) {
        return GroupEF.width(components);
      } else if(right.equals("height") || right.equals("y")) {
        return GroupEF.height(components);
      } else if(right.equals("left")) {
        return GroupEF.left(components);
      } else if(right.equals("top")) {
        return GroupEF.top(components);
      } else if(right.equals("heightmax")) {
        return GroupEF.heightMax(components);
      } else if(right.equals("heightmin")) {
        return GroupEF.heightMin(components);
      } else if(right.equals("heightsum")) {
        return GroupEF.heightSum(components);
      } else if(right.equals("widthmax")) {
        return GroupEF.widthMax(components);
      } else if(right.equals("widthmin")) {
        return GroupEF.widthMin(components);
      } else if(right.equals("widthsum")) {
        return GroupEF.widthSum(components);
      } else if(right.equals("prefheightmax")) {
        return GroupEF.preferredHeightMax(components);
      } else if(right.equals("prefheightmin")) {
        return GroupEF.preferredHeightMin(components);
      } else if(right.equals("prefheightsum")) {
        return GroupEF.preferredHeightSum(components);
      } else if(right.equals("prefwidthmax")) {
        return GroupEF.preferredWidthMax(components);
      } else if(right.equals("prefwidthmin")) {
        return GroupEF.preferredWidthMin(components);
      } else if(right.equals("prefwidthsum")) {
        return GroupEF.preferredWidthSum(components);
      }
      throw new IllegalStateException("Unknown constraint \"" + token + "\" for the group of mapped levels \"" + left + "\"!");
    }
    Component component = components[0];
    if(right.equals("bottom")) {
      if(component == parentContainer) {
        return ContainerEF.bottom(parentContainer);
      }
      return ComponentEF.bottom(component);
    } else if(right.equals("right")) {
      if(component == parentContainer) {
        return ContainerEF.right(parentContainer);
      }
      return ComponentEF.right(component);
    } else if(right.equals("width") || right.equals("x")) {
      if(component == parentContainer) {
        return ContainerEF.width(parentContainer);
      }
      return ComponentEF.width(component);
    } else if(right.equals("height") || right.equals("y")) {
      if(component == parentContainer) {
        return ContainerEF.height(parentContainer);
      }
      return ComponentEF.height(component);
    } else if(right.equals("left")) {
      if(component == parentContainer) {
        return ContainerEF.left(parentContainer);
      }
      return ComponentEF.left(component);
    } else if(right.equals("top")) {
      if(component == parentContainer) {
        return ContainerEF.top(parentContainer);
      }
      return ComponentEF.top(component);
    } else if(right.equals("maxheight")) {
      return ComponentEF.maximumHeight(component);
    } else if(right.equals("minheight")) {
      return ComponentEF.minimumHeight(component);
    } else if(right.equals("maxwidth")) {
      return ComponentEF.maximumWidth(component);
    } else if(right.equals("minwidth")) {
      return ComponentEF.minimumWidth(component);
    } else if(right.equals("prefheight")) {
      return ComponentEF.preferredHeight(component);
    } else if(right.equals("prefwidth")) {
      return ComponentEF.preferredWidth(component);
    }
    throw new IllegalStateException("Unknown constraint \"" + token + "\" for the level \"" + left + "!");
  }

}
