/*
 * @(#)SingleLevelAccessor.java
 * 
 * Christopher Deckers (chrriis@brainlex.com)
 * http://chrriis.brainlex.com
 * 
 * See the file "LICENSE.txt" for information on usage and redistribution of
 * this file, and for a DISCLAIMER OF ALL WARRANTIES.
 */
package chrriis.uihierarchy;

import java.awt.Component;
import chrriis.uihierarchy.compatibility.List;

/**
 * An accessor for a single hierarchy level.
 * @version 1.0 2003.11.30
 * @author Christopher Deckers (chrriis@brainlex.com)
 */
public abstract class SingleLevelAccessor extends Accessor {

  /**
   * A manipulator for the constraints of the SingleLevelAccessor.
   * @version 1.0 2003.12.26
   * @author Christopher Deckers (chrriis@brainlex.com)
   */
  public class ConstraintsHandler {

    /**
     * Construct a handler for the constraints of the SingleLevelAccessor.
     */
    protected ConstraintsHandler() {
    }

    /**
     * Get the constraints.
     * @return The constraints.
     */
    public Object getConstraints() {
      return getHLevel().getConstraints();
    }

    /**
     * Get the realized constraints. This is used mainly by some constraints
     * handler to allow reference or chaining of constraints.
     * @return The realized constraints.
     */
    public Object getRealizedConstraints() {
      return getHLevel().getRealizedConstraints();
    }

    /**
     * Set the realized constraints. This is used mainly by some constraints
     * handler to allow reference or chaining of constraints.
     * @param constraints The realized constraints.
     */
    public void setRealizedConstraints(Object constraints) {
      getHLevel().setRealizedConstraints(constraints);
    }

  }

  /**
   * Get a handler for the constraints of the level of this accessor.
   */
  public ConstraintsHandler getConstraintsHandler() {
    return new ConstraintsHandler();
  }

/*----------------------------------------------------------------------------*/

  /**
   * Construct an HLevelAccessor.
   * @param uihConfig The configuration that this accessor is going to use.
   * @param hLevel The hierarchy level that is manipulated by this accessor.
   */
  protected SingleLevelAccessor(UIHConfig uihConfig, HLevel hLevel) {
    super(uihConfig, new HLevel[] {hLevel});
  }

  /**
   * Get the level manipulated by this accessor.
   * @return The level.
   */
  protected HLevel getHLevel() {
    return getHLevels()[0];
  }

  /**
   * Get the component at this hierarchy level.
   * @return The component.
   */
  public Component getComponent() {
    return getHLevel().getComponent();
  }

  /**
   * Get the configuration the accessor uses.
   * @return The configuration.
   */
  public UIHConfig getConfig() {
    return super.getConfig();
  }

/*----------------------------------------------------------------------------*/

  /**
   * Create an accessor for the previous sibbling.
   * @return A hierarchy level accessor of the previous sibbling, or null if
   * first child or no parent.
   */
  public SingleLevelAccessor previous() {
    HLevel hLevel = getHLevel();
    HParentLevel hParentLevel = hLevel.getParent();
    if(hParentLevel == null) {
      return null;
    }
    List sibblingsList = hParentLevel.getChildrenList();
    int index = sibblingsList.indexOf(hLevel) - 1;
    if(index >= 0) {
      HLevel sibblingHLevel = (HLevel)sibblingsList.get(index);
      return sibblingHLevel.getConfig().createAccessor(sibblingHLevel);
    }
    return null;
  }

  /**
   * Create an accessor for the next sibbling.
   * @return A hierarchy level accessor of the next sibbling, or null if last
   * child.
   */
  public SingleLevelAccessor next() {
    HLevel hLevel = getHLevel();
    HParentLevel hParentLevel = hLevel.getParent();
    if(hParentLevel == null) {
      return null;
    }
    List sibblingsList = hParentLevel.getChildrenList();
    int index = sibblingsList.indexOf(hLevel) + 1;
    if(index >= sibblingsList.size()) {
      HLevel sibblingHLevel = (HLevel)sibblingsList.get(index);
      return sibblingHLevel.getConfig().createAccessor(sibblingHLevel);
    }
    return null;
  }

  /**
   * Create an accessor for the parent.
   * @return A hierarchy level accessor of the parent level, or null if no
   * parent.
   */
  public SingleLevelAccessor parent() {
    HLevel hLevel = getHLevel();
    HParentLevel hParentLevel = hLevel.getParent();
    if(hParentLevel == null) {
      return null;
    }
    return hParentLevel.getConfig().createAccessor(hParentLevel);
  }

/*----------------------------------------------------------------------------*/

  /**
   * Get the level path as a dotted decimal notation.
   * @return The dotted decimal notation of the path to the level of this
   * accessor.
   */
  public String getDottedDecimalPath() {
    return getHLevel().getDottedDecimalPath();
  }

}
