/*
 * @(#)AccessorFactory.java
 * 
 * Christopher Deckers (chrriis@brainlex.com)
 * http://chrriis.brainlex.com
 * 
 * See the file "LICENSE.txt" for information on usage and redistribution of
 * this file, and for a DISCLAIMER OF ALL WARRANTIES.
 */
package chrriis.uihierarchy.accessor;

import chrriis.uihierarchy.Accessor;
import chrriis.uihierarchy.HLevel;
import chrriis.uihierarchy.HParentLevel;
import chrriis.uihierarchy.ParentLevelAccessor;
import chrriis.uihierarchy.SingleLevelAccessor;
import chrriis.uihierarchy.UIHConfig;

/**
 * The factory that is used to create the accessors.
 * @version 1.0 2003.12.2
 * @author Christopher Deckers (chrriis@brainlex.com)
 */
public class AccessorFactory {

  /**
   * Construct an AccessorFactory. This class can be subclassed to provide
   * custom accessors with additional functionalities.
   */
  public AccessorFactory() {}

  /**
   * Create an accessor for some levels of a hierarchy. This method is final as
   * it performs some checks on the accessors that are created. If you want to
   * provide another factory, then you should override
   * <code>newAccessor(UIHConfig, HLevel[])</code> which is called when more
   * than one level is to be handled. When only one level is concerned
   * <code>newSingleLevelAccessor(UIHConfig, HLevel)</code> is invoked.
   * @param uihConfig The configuration that this accessor is going to use.
   * @param hLevels The levels for which to create the accessor.
   * @return The accessor of the levels.
   */
  public final Accessor createAccessor(UIHConfig uihConfig, HLevel[] hLevels) {
    if(hLevels.length == 1) {
      return newSingleLevelAccessor(uihConfig, hLevels[0]);
    }
    return newAccessor(uihConfig, hLevels);
  }

  /**
   * Create an accessor for a single parent level of a hierarchy. It is called
   * by <code>newSingleLevelAccessor(...)</code> to create an accessor that
   * manipulates a single parent level. This method may be overriden when
   * defining a new factory.
   * @param uihConfig The configuration that this accessor is going to use.
   * @param hParentLevel The parent level for which to create the accessor.
   * @return The accessor of the parent level.
   */
  public ParentLevelAccessor newParentLevelAccessor(UIHConfig uihConfig, HParentLevel hParentLevel) {
    // Subclass to access constructor. I don't want (yet) the constructor to be
    // public as people may be tempted to keep references on levels...
    return new ParentLevelAccessor(uihConfig, hParentLevel) {};
  }

  /**
   * Create an accessor for a single level of a hierarchy. It is called by the
   * framework to create an accessor that manipulates a single level. This
   * method may be overriden when defining a new factory. Unless overriden, this
   * method calls <code>newParentLevelAccessor(...)</code> when the level is a
   * parent level.
   * @param uihConfig The configuration that this accessor is going to use.
   * @param hLevel The level for which to create the accessor.
   * @return The accessor of the level.
   */
  public SingleLevelAccessor newSingleLevelAccessor(UIHConfig uihConfig, HLevel hLevel) {
    // Subclass to access constructor. I don't want (yet) the constructor to be
    // public as people may be tempted to keep references on levels...
    if(hLevel instanceof HParentLevel) {
      return newParentLevelAccessor(uihConfig, (HParentLevel)hLevel);
    }
    return new SingleLevelAccessor(uihConfig, hLevel) {};
  }

  /**
   * Create an accessor for some levels of a hierarchy. It is called by the
   * framework to create an accessor that manipulates more than one level at a
   * time. This method may be overriden when defining a new factory.
   * @param uihConfig The hierarchy configuration that this accessor is going
   * to act on.
   * @param hLevels The levels for which to create the accessor.
   * @return The accessor of the levels.
   */
  public Accessor newAccessor(UIHConfig uihConfig, HLevel[] hLevels) {
    // Subclass to access constructor. I don't want (yet) the constructor to be
    // public as people may be tempted to keep references on levels...
    return new Accessor(uihConfig, hLevels) {};
  }

}
